from guillotina.interfaces import IContainer
from typing import Awaitable
from zope.interface import Interface


class IEvolutionUtility(Interface):
    def register(self, version: int, evolver: Awaitable):
        pass

    async def evolve(self, container: IContainer):
        pass



from guillotina.component import get_utility
from guillotina.content import create_content_in_container
from guillotina.transactions import transaction
from guillotina.utils import get_registry
from guillotina_evolution.tests.fixtures import ctx
from guillotina_evolution.utility import GENERATION_KEY
from guillotina_evolution.utility import IEvolutionUtility
from guillotina_evolution.utils import register_evolution


async def test_multiple_evolutions_at_once(loop, my_requester):
    async with ctx(my_requester) as container:
        # Create content in the container
        async with transaction():
            await create_content_in_container(container, "Item", id_="foobar")
            ob = await container.async_get("foobar")
            assert hasattr(ob, "title") is False

        # Register a new evolution
        async def ensure_all_items_have_attribute_title(container):
            async for item in container.async_values():
                item.title = ""
                item.register()

        async def ensure_all_items_have_attribute_description(container):
            async for item in container.async_values():
                item.description = "patata"
                item.register()

        utility = get_utility(IEvolutionUtility)

        async with transaction():
            await utility._update_curr_gen(0)

        utility.register(1, ensure_all_items_have_attribute_title)
        utility.register(2, ensure_all_items_have_attribute_description)

        # Evolve
        await utility.evolve(container)

        # Assert generation was updated on container registry
        async with transaction():
            registry = await get_registry()
            assert registry[GENERATION_KEY] == 2

        # Check objects after evolution
        async with transaction():
            ob = await container.async_get("foobar")
            assert ob is not None
            assert hasattr(ob, "title") is True
            assert hasattr(ob, "description") is True
            assert ob.title == ""
            assert ob.description == "patata"


async def test_registry(my_requester):
    async with ctx(my_requester):
        utility = get_utility(IEvolutionUtility)

        async with transaction():
            assert await utility._get_curr_gen() == 0

        async with transaction():
            await utility._update_curr_gen(5)

        registry = await get_registry()
        assert registry[GENERATION_KEY] == 5


def test_register_decorator():
    @register_evolution(1)
    async def migration_step(container):
        pass

    utility = get_utility(IEvolutionUtility)
    assert utility._evolvers == {1: migration_step}



from guillotina.component import get_utility
from guillotina_evolution.interfaces import IEvolutionUtility


def register_evolution(gen):
    def decorator(evolver):
        utility = get_utility(IEvolutionUtility)
        utility.register(gen, evolver)
        return evolver

    return decorator



from guillotina.commands import Command
from guillotina.component import get_utility
from guillotina.tests.utils import login
from guillotina.utils import get_containers
from guillotina_evolution.utility import IEvolutionUtility


class EvolveCommand(Command):
    description = "Run evolution"

    async def run(self, arguments, settings, app):
        login()
        utility = get_utility(IEvolutionUtility)
        async for _, _, container in get_containers():
            await utility.evolve(container)



from guillotina import configure
from guillotina.interfaces import IContainer
from guillotina.transactions import transaction
from guillotina.utils import get_registry
from guillotina_evolution.interfaces import IEvolutionUtility
from typing import Awaitable

import logging

logger = logging.getLogger("guillotina_evolution")


GENERATION_KEY = "guillotina_evolution.utility.generation"


@configure.utility(provides=IEvolutionUtility)
class EvolutionUtility(object):
    def __init__(self, settings=None, loop=None):
        self._settings = settings or {}
        self._loop = loop
        self._evolvers = {}

    async def install(self):
        async with transaction():
            await self._init_registry()

    def register(self, gen: int, evolver: Awaitable):
        if gen in self._evolvers:
            raise Exception(f"Evolver for generation '{gen}' already exist")

        self._evolvers[gen] = evolver

    async def evolve(self, container: IContainer):
        async with transaction():
            cur_gen = await self._get_curr_gen()
            evolvers = self._get_evolvers(cur_gen)

        if len(evolvers) > 0:
            logger.info(f"Start evolving container {container}")
            for gen, evolver in evolvers:
                async with transaction():
                    logger.info(f"Evolving from generation '{cur_gen}' to '{gen}'")
                    await evolver(container)
                    cur_gen = await self._update_curr_gen(gen)

            logger.info(f"Container {container} is now at generation {gen}")
        else:
            logger.info(f"Container already at latest generation")

    def _get_evolvers(self, from_):
        evolvers = []
        for gen in sorted(self._evolvers.keys()):
            if gen > from_:
                evolvers += [(gen, self._evolvers[gen])]
        return evolvers

    async def _get_curr_gen(self):
        registry = await get_registry()
        return registry[GENERATION_KEY]

    async def _update_curr_gen(self, gen):
        registry = await get_registry()
        registry[GENERATION_KEY] = gen
        registry.register()
        return gen

    async def _init_registry(self):
        registry = await get_registry()
        registry[GENERATION_KEY] = self._get_greatest_registered_gen()
        registry.register()

    def _get_greatest_registered_gen(self):
        gens = sorted(self._evolvers.keys())
        if len(gens) > 0:
            return gens[-1]
        else:
            return 0



from guillotina import configure

app_settings = {
    # provide custom application settings here...
}


def includeme(root):
    """
    custom application initialization here
    """
    configure.scan("guillotina_evolution.commands")
    configure.scan("guillotina_evolution.utility")



from setuptools import find_packages
from setuptools import setup


try:
    README = open('README.rst').read()
except IOError:
    README = None

setup(
    name='guillotina_evolution',
    version="1.0.0",
    description=' ',
    long_description=README,
    install_requires=[
        'guillotina>=5.0,<6.0',
    ],
    author='Jordi Masip',
    author_email='jmasip@vinissimus.com',
    url='https://github.com/vinissimus/guillotina_evolution',
    packages=find_packages(exclude=['demo']),
    include_package_data=True,
    tests_require=[
        'pytest',
    ],
    extras_require={
        'test': [
            'pytest',
            'pytest-asyncio',
            'pytest-cov',
            'coverage',
        ]
    },
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Natural Language :: English",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ],
    entry_points={
    }
)



from contextlib import asynccontextmanager
from guillotina import task_vars
from guillotina import testing
from guillotina.component import get_utility
from guillotina.tests import utils
from guillotina.tests.fixtures import ContainerRequesterAsyncContextManager
from guillotina.tests.utils import get_container
from guillotina_evolution.interfaces import IEvolutionUtility

import json
import pytest


def base_settings_configurator(settings):
    if "applications" in settings:
        settings["applications"].append("guillotina_evolution")
    else:
        settings["applications"] = ["guillotina_evolution"]


testing.configure_with(base_settings_configurator)


class guillotina_evolution_Requester(ContainerRequesterAsyncContextManager):  # noqa
    async def __aenter__(self):
        await super().__aenter__()
        await self.requester(
            "POST",
            "/db/guillotina/@addons",
            data=json.dumps({"id": "guillotina_evolution"}),
        )
        utility = get_utility(IEvolutionUtility)
        await get_container(db=self.requester.db)
        await utility.install()
        return self.requester


@pytest.fixture(scope="function")
async def my_requester(guillotina):
    return guillotina_evolution_Requester(guillotina)


@asynccontextmanager
async def ctx(my_requester):
    async with my_requester as requester:
        task_vars.db.set(requester.db)
        container = await get_container(db=requester.db)
        utils.login()
        yield container
